/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.UUID;

import org.openepics.names.docker.ITUtil.AuthorizationChoice;
import org.openepics.names.docker.ITUtil.EndpointChoice;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.response.ResponseBoolean;
import org.openepics.names.rest.beans.response.ResponseBooleanList;
import org.openepics.names.rest.beans.response.ResponsePageStructureElements;
import org.openepics.names.util.EnumUtil.StructureChoice;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Utility class to help (Docker) integration tests for Naming and PostgreSQL.
 *
 * @author Lars Johansson
 */
public class ITUtilStructureElement {

    static final ObjectMapper mapper = new ObjectMapper();

    /**
     * This class is not to be instantiated.
     */
    private ITUtilStructureElement() {
        throw new IllegalStateException("Utility class");
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return string for structure element array.
     *
     * @param value structure element array
     * @return string for structure element array
     */
    static String object2Json(StructureElementCommand[] value) {
        try {
            return mapper.writeValueAsString(value);
        } catch (JsonProcessingException e) {
            fail();
        }
        return null;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that array of structure elements has expected length.
     *
     * @param actual array of structure elements
     * @param expectedLength expected length of array
     */
    static void assertContentLength(StructureElement[] actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.length);
    }

    /**
     * Assert that array of structure elements has length one and structure element is available.
     *
     * @param actual array of structure elements
     * @return structure element
     */
    static StructureElement assertContentLengthOne(StructureElement[] actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual[0]);
        return actual[0];
    }

    /**
     * Assert that response page with list of structure elements has expected length.
     *
     * @param actual response page with list of structure elements
     * @param expectedLength expected length of list
     */
    static void assertContentLength(ResponsePageStructureElements actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.getList().size());
        assertEquals(expectedLength, actual.getListSize());
    }

    /**
     * Assert that response page with list of structure elements has length one and structure element is available.
     *
     * @param actual response page with list of structure elements
     * @return structure element
     */
    static StructureElement assertContentLengthOne(ResponsePageStructureElements actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual.getList().get(0));
        return actual.getList().get(0);
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that arrays are equal with same length and same content in each array position.
     *
     * @param actual actual array of structure elements
     * @param expected expected arbitrary number of structure elements
     */
    static void assertEqualsStructureElements(StructureElement[] actual, StructureElement... expected) {
        if (expected != null) {
            assertNotNull(actual);
            assertEquals(expected.length, actual.length);
            for (int i=0; i<expected.length; i++) {
                assertTrue(expected[i].equals(actual[i]));
            }
        } else {
            assertNull(actual);
        }
    }

    /**
     * Assert content of structure element.
     *
     * @param actual structure element
     * @param expectedType expected type
     * @param expectedUuid expected uuid
     * @param expectedParent expected parent
     * @param expectedName expected name
     * @param expectedMnemonic expected mnemonic
     * @param expectedMnemonicpath expected mnemonic path
     * @param expectedLevel expected level
     * @param expectedDescription expected description
     * @param expectedStatus expected status
     * @param expectedIsLatest expected latest
     * @param expectedIsDeleted expected deleted
     * @param expectedWho expected who
     * @param expectedComment expected comment
     */
    static void assertContent(StructureElement actual,
            Type expectedType, UUID expectedUuid, UUID expectedParent,
            String expectedName, String expectedMnemonic, String expectedMnemonicpath, Integer expectedLevel,
            String expectedDescription, Status expectedStatus, Boolean expectedIsLatest, Boolean expectedIsDeleted,
            String expectedWho, String expectedComment) {
        assertNotNull(actual);
        assertEquals(expectedType, actual.getType());
        assertEquals(expectedUuid, actual.getUuid());
        assertEquals(expectedParent, actual.getParent());
        assertEquals(expectedName, actual.getName());
        assertEquals(expectedMnemonic, actual.getMnemonic());
        assertEquals(expectedMnemonicpath, actual.getMnemonicpath());
        assertEquals(expectedLevel, actual.getLevel());
        assertEquals(expectedDescription, actual.getDescription());
        assertEquals(expectedStatus, actual.getStatus());
        assertEquals(expectedIsLatest, actual.isLatest());
        assertEquals(expectedIsDeleted, actual.isDeleted());
        assertNotNull(actual.getWhen());
        assertEquals(expectedWho, actual.getWho());
        assertEquals(expectedComment, actual.getComment());
    }

    /**
     * Assert content of structure element.
     *
     * @param actual structure element
     * @param expected structure element command
     */
    static void assertContent(StructureElement actual, StructureElementCommand expected) {
        assertNotNull(actual);
        if (expected != null) {
            assertEquals(expected.getUuid(), actual.getUuid());
            assertEquals(expected.getType(), actual.getType());
            assertEquals(expected.getParent(), actual.getParent());
            assertEquals(expected.getName(), actual.getName());
            assertEquals(expected.getMnemonic(), actual.getMnemonic());
            assertEquals(expected.getDescription(), actual.getDescription());
            assertEquals(expected.getComment(), actual.getComment());
        }
    }


    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to read structure elements.
     *
     * @param queryString query string
     * @return response page structure elements
     */
    public static ResponsePageStructureElements assertRead(String queryString) {
        return assertRead(queryString, -1, -1);
    }
    /**
     * Utility method to read structure elements and assert number of items in list.
     *
     * @param queryString query string
     * @param expectedEqual (if non-negative number) equal to this number of items
     * @return response page structure elements
     */
    public static ResponsePageStructureElements assertRead(String queryString, int expectedEqual) {
        return assertRead(queryString, expectedEqual, expectedEqual);
    }
    /**
     * Utility method to read structure elements and assert number of items in list.
     *
     * @param queryString query string
     * @param expectedGreaterThanOrEqual (if non-negative number) greater than or equal to this number of items
     * @param expectedLessThanOrEqual (if non-negative number) less than or equal to this number of items
     * @return response page structure elements
     */
    public static ResponsePageStructureElements assertRead(String queryString, int expectedGreaterThanOrEqual, int expectedLessThanOrEqual) {
        try {
            String[] response = null;
            ResponsePageStructureElements responsePageStructureElements = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + queryString);
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            responsePageStructureElements = mapper.readValue(response[1], ResponsePageStructureElements.class);

            // expected number of items in list
            //     (if non-negative number)
            //     expectedGreaterThanOrEqual <= nbr of items <= expectedLessThanOrEqual
            if (expectedGreaterThanOrEqual >= 0) {
                assertTrue(responsePageStructureElements.getListSize() >= expectedGreaterThanOrEqual);
            }
            if (expectedLessThanOrEqual >= 0) {
                assertTrue(responsePageStructureElements.getListSize() <= expectedLessThanOrEqual);
            }

            // expected value for latest for items in list
            assertLatest(responsePageStructureElements);

            return responsePageStructureElements;
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
        return null;
    }

    /**
     * Assert that response page with list of structure elements has maximum one entry that is latest
     * for any given uuid, which may not have status {@link Status#PENDING}.
     *
     * @param actual response page with list of structure elements
     */
    public static void assertLatest(ResponsePageStructureElements actual) {
        // latest
        //     status pending not with latest
        //     maximum one item for any given uuid in list may be latest

        assertNotNull(actual);
        Map<UUID, Long> mapUuidCountLatest = new TreeMap<>();
        for (StructureElement structureElement : actual.getList()) {
            if (structureElement.isLatest()) {
                ITUtil.addOne(structureElement.getUuid(), mapUuidCountLatest);
                assertNotEquals(Status.PENDING, structureElement.getStatus());
            }
        }
        for (Entry<UUID, Long> entry : mapUuidCountLatest.entrySet()) {
            assertTrue(entry.getValue() <= 1);
        }
    }

    /**
     * Utility method to get name equivalence of a mnemonic or name.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertEquivalence(String name, String expected) {
        try {
            String[] response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/equivalence/" + name);
            ITUtil.assertResponseLength2CodeOKContent(response, expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if structure exists.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     */
    public static void assertExists(Type type, String mnemonicpath, Boolean expected) {
        assertExists(type, mnemonicpath, expected, null);
    }
    /**
     * Utility method to check if structure exists.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     * @param expectedMessageEmpty expected message empty
     */
    public static void assertExists(Type type, String mnemonicpath, Boolean expected, Boolean expectedMessageEmpty) {
        try {
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/exists/" + type.toString() + "/" + mnemonicpath);
            ITUtil.assertResponseLength2CodeOK(response);
            if (expectedMessageEmpty != null) {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected, expectedMessageEmpty);
            } else {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
            }
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if structure is valid to create.
     * Method will not produce a message.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     */
    public static void assertIsValidToCreate(Type type, String mnemonicpath, Boolean expected) {
        try {
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/isvalidtocreate/" + type.toString() + "/" + mnemonicpath);
            ITUtil.assertResponseLength2CodeOK(response);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to (try to) validate a structure element from given json and assert expected response.
     * Method expected to NOT succeed to validate
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param json json
     * @param structureChoice structure choice
     * @param responseCode response code
     */
    public static void assertValidate(String json, StructureChoice structureChoice, int responseCode) {
        assertValidate(AuthorizationChoice.NONE, json, structureChoice, responseCode);
    }
    /**
     * Utility method to (try to) validate a structure element from given json and assert expected response.
     * Method expected to NOT succeed to validate
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param structureChoice structure choice
     * @param responseCode response code
     */
    public static void assertValidate(AuthorizationChoice authorizationChoice, String json, StructureChoice structureChoice, int responseCode) {
        String validatePath = getValidatePath(structureChoice);

        try {
            String[] response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, validatePath, json));
            ITUtil.assertResponseLength2Code(response, responseCode);
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    // --------------------

    /**
     * Utility method to validate structure element and assert expected response.
     *
     * @param structureElement structure element
     * @param structureChoice structure choice
     * @param expected expected response
     */
    public static void assertValidate(StructureElementCommand structureElement, StructureChoice structureChoice, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement}, structureChoice, expected);
    }
    /**
     * Utility method to validate structure elements and assert expected response.
     *
     * @param structureElements structure elements
     * @param structureChoice structure choice
     * @param expected expected response
     */
    public static void assertValidate(StructureElementCommand[] structureElements, StructureChoice structureChoice, Boolean expected) {
        assertValidate(AuthorizationChoice.NONE, structureElements, structureChoice, expected);
    }
    /**
     * Utility method to validate structure elements and assert expected response.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @param structureChoice structure choice
     * @param expected expected response
     */
    public static void assertValidate(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements, StructureChoice structureChoice, Boolean expected) {
        String path = getValidatePath(structureChoice);

        try {
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), expected);
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to return validate path for structure choice.
     *
     * @param structureChoice structure choice
     * @return validate path
     */
    private static String getValidatePath(StructureChoice structureChoice) {
        String validatePath = "";
        switch (structureChoice) {
            case CREATE:
                validatePath = "/validatecreate";
                break;
            case UPDATE:
                validatePath = "/validateupdate";
                break;
            case DELETE:
                validatePath = "/validatedelete";
                break;
            case APPROVE:
                validatePath = "/validateapprove";
                break;
            case CANCEL:
                validatePath = "/validatecancel";
                break;
            case REJECT:
                validatePath = "/validatereject";
                break;
            default:
                break;
        }
        return validatePath;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to (try to) create a structure element and assert result response code.
     * Method expected to NOT succeed to create a structure element
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param structureElement structure element
     * @param responseCode response code
     * @throws JsonProcessingException
     */
    public static void assertCreate(StructureElementCommand structureElement, int responseCode) throws JsonProcessingException {
        assertCreate(AuthorizationChoice.NONE, object2Json(new StructureElementCommand[] {structureElement}), responseCode);
    }
    /**
     * Utility method to (try to) create a structure element from given json and assert result response code.
     * Method expected to NOT succeed to create a structure element
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param json json
     * @param responseCode response code
     */
    public static void assertCreate(String json, int responseCode) {
        assertCreate(AuthorizationChoice.NONE, json, responseCode);
    }
    /**
     * Utility method to (try to) create a structure element from given json and assert result response code.
     * Method expected to NOT succeed to create a structure element
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param json json
     * @param responseCode response code
     */
    public static void assertCreate(AuthorizationChoice authorizationChoice, String json, int responseCode) {
        try {
            String[] response = ITUtil.runShellCommand(ITUtil.curlPostPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "", json));
            ITUtil.assertResponseLength2Code(response, responseCode);
            // ITUtil.assertMessageNotEmpty(mapper.readValue(response[1], Response.class));
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    // --------------------

    /**
     * Utility method to create a structure element and assert result.
     *
     * @param structureElement structure element
     * @return created structure element
     */
    public static StructureElement assertCreate(StructureElementCommand structureElement) {
        return assertCreate(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to create structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return created structure element
     */
    public static StructureElement[] assertCreate(StructureElementCommand[] structureElements) {
        return assertCreate(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to create structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return created structure element
     */
    public static StructureElement[] assertCreate(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.CREATE);

        try {
            String[] response = null;
            StructureElement[] createdStructureElements  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPostPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_CREATED);
            createdStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement createdStructureElement : createdStructureElements) {
                assertNotNull(createdStructureElement.getUuid());
                assertEquals(Status.PENDING, createdStructureElement.getStatus());
                assertEquals(Boolean.FALSE, createdStructureElement.isLatest());
                assertEquals(Boolean.FALSE, createdStructureElement.isDeleted());
                assertNotNull(createdStructureElement.getWhen());
            }

            // not set uuid for input parameter(s)
            //     convenience, easy when array length 1, otherwise not easy

            return createdStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to update a structure element and assert result.
     *
     * @param structureElement structure element
     * @return updated structure element
     */
    public static StructureElement assertUpdate(StructureElementCommand structureElement) {
        return assertUpdate(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to update structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return updated structure element
     */
    public static StructureElement[] assertUpdate(StructureElementCommand[] structureElements) {
        return assertUpdate(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to update structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return updated structure element
     */
    public static StructureElement[] assertUpdate(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.UPDATE);

        try {
            String[] response = null;
            StructureElement[] updatedStructureElements  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPutPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            updatedStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement updatedStructureElement : updatedStructureElements) {
                assertNotNull(updatedStructureElement.getUuid());
                assertEquals(Status.PENDING, updatedStructureElement.getStatus());
                assertEquals(Boolean.FALSE, updatedStructureElement.isLatest());
                assertEquals(Boolean.FALSE, updatedStructureElement.isDeleted());
                assertNotNull(updatedStructureElement.getWhen());
            }

            return updatedStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to delete a structure element and assert result.
     *
     * @param structureElement structure element
     * @return deleted structure element
     */
    public static StructureElement assertDelete(StructureElementCommand structureElement) {
        return assertDelete(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to delete structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return deleted structure element
     */
    public static StructureElement[] assertDelete(StructureElementCommand[] structureElements) {
        return assertDelete(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to delete structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return deleted structure element
     */
    public static StructureElement[] assertDelete(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.DELETE);

        try {
            String[] response = null;
            StructureElement[] deletedStructureElements  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlDeletePathJson(authorizationChoice, EndpointChoice.STRUCTURES, "", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            deletedStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement deletedStructureElement : deletedStructureElements) {
                assertNotNull(deletedStructureElement.getUuid());
                assertEquals(Status.PENDING, deletedStructureElement.getStatus());
                assertEquals(Boolean.FALSE, deletedStructureElement.isLatest());
                assertEquals(Boolean.TRUE, deletedStructureElement.isDeleted());
                assertNotNull(deletedStructureElement.getWhen());
            }

            return deletedStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element
     */
    public static StructureElement assertApprove(StructureElementCommand structureElement) {
        return assertApprove(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to approve structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return approved structure element
     */
    public static StructureElement[] assertApprove(StructureElementCommand[] structureElements) {
        return assertApprove(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to approve structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return approved structure element
     */
    public static StructureElement[] assertApprove(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.APPROVE);

        try {
            String[] response = null;
            StructureElement[] approvedStructureElements = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "/approve", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            approvedStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement approvedStructureElement : approvedStructureElements) {
                assertNotNull(approvedStructureElement.getUuid());
                assertEquals(Status.APPROVED, approvedStructureElement.getStatus());
                assertEquals(Boolean.TRUE, approvedStructureElement.isLatest());
                // not known if deleted
                assertNotNull(approvedStructureElement.getWhen());
            }

            return approvedStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element
     */
    public static StructureElement assertCancel(StructureElementCommand structureElement) {
        return assertCancel(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to cancel structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return cancelled structure element
     */
    public static StructureElement[] assertCancel(StructureElementCommand[] structureElements) {
        return assertCancel(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to cancel structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return cancelled structure element
     */
    public static StructureElement[] assertCancel(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.CANCEL);

        try {
            String[] response = null;
            StructureElement[] cancelledStructureElements = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "/cancel", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            cancelledStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement cancelledStructureElement : cancelledStructureElements) {
                assertNotNull(cancelledStructureElement.getUuid());
                assertEquals(Status.CANCELLED, cancelledStructureElement.getStatus());
                assertEquals(Boolean.FALSE, cancelledStructureElement.isLatest());
                // not known if deleted
                assertNotNull(cancelledStructureElement.getWhen());
            }

            return cancelledStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element
     */
    public static StructureElement assertReject(StructureElementCommand structureElement) {
        return assertReject(AuthorizationChoice.NONE, new StructureElementCommand[] {structureElement})[0];
    }
    /**
     * Utility method to reject structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return rejected structure element
     */
    public static StructureElement[] assertReject(StructureElementCommand[] structureElements) {
        return assertReject(AuthorizationChoice.NONE, structureElements);
    }
    /**
     * Utility method to reject structure elements and assert result.
     *
     * @param authorizationChoice authorization choice (none, user, admin)
     * @param structureElements structure elements
     * @return rejected structure element
     */
    public static StructureElement[] assertReject(AuthorizationChoice authorizationChoice, StructureElementCommand[] structureElements) {
        String path = getValidatePath(StructureChoice.REJECT);

        try {
            String[] response = null;
            StructureElement[] rejectedStructureElements = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(authorizationChoice, EndpointChoice.STRUCTURES, path, object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(authorizationChoice, EndpointChoice.STRUCTURES, "/reject", object2Json(structureElements)));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            rejectedStructureElements = mapper.readValue(response[1], StructureElement[].class);

            for (StructureElement rejectedStructureElement : rejectedStructureElements) {
                assertNotNull(rejectedStructureElement.getUuid());
                assertEquals(Status.REJECTED, rejectedStructureElement.getStatus());
                assertEquals(Boolean.FALSE, rejectedStructureElement.isLatest());
                // not known if deleted
                assertNotNull(rejectedStructureElement.getWhen());
            }

            return rejectedStructureElements;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    // --------------------

    /**
     * Utility method to create, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after create)
     */
    public static StructureElement assertCreateApprove(StructureElementCommand structureElement) {
        return assertApprove(assertCreate(structureElement));
    }
    /**
     * Utility method to create, approve structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return approved structure elements (after create)
     */
    public static StructureElement[] assertCreateApprove(StructureElementCommand[] structureElements) {
        return assertApprove(assertCreate(structureElements));
    }

    /**
     * Utility method to create, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after create)
     */
    public static StructureElement assertCreateCancel(StructureElementCommand structureElement) {
        return assertCancel(assertCreate(structureElement));
    }
    /**
     * Utility method to create, cancel structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return cancelled structure elements (after create)
     */
    public static StructureElement[] assertCreateCancel(StructureElementCommand[] structureElements) {
        return assertCancel(assertCreate(structureElements));
    }

    /**
     * Utility method to create, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after create)
     */
    public static StructureElement assertCreateReject(StructureElementCommand structureElement) {
        return assertReject(assertCreate(structureElement));
    }
    /**
     * Utility method to create, reject structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return rejected structure elements (after create)
     */
    public static StructureElement[] assertCreateReject(StructureElementCommand[] structureElements) {
        return assertReject(assertCreate(structureElements));
    }

    /**
     * Utility method to update, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after update)
     */
    public static StructureElement assertUpdateApprove(StructureElementCommand structureElement) {
        return assertApprove(assertUpdate(structureElement));
    }
    /**
     * Utility method to update, approve structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return approved structure elements (after update)
     */
    public static StructureElement[] assertUpdateApprove(StructureElementCommand[] structureElements) {
        return assertApprove(assertUpdate(structureElements));
    }

    /**
     * Utility method to update, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after update)
     */
    public static StructureElement assertUpdateCancel(StructureElementCommand structureElement) {
        return assertCancel(assertUpdate(structureElement));
    }
    /**
     * Utility method to update, cancel structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return cancelled structure elements (after update)
     */
    public static StructureElement[] assertUpdateCancel(StructureElementCommand[] structureElements) {
        return assertCancel(assertUpdate(structureElements));
    }

    /**
     * Utility method to update, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after update)
     */
    public static StructureElement assertUpdateReject(StructureElementCommand structureElement) {
        return assertReject(assertUpdate(structureElement));
    }
    /**
     * Utility method to update, reject structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return rejected structure elements (after update)
     */
    public static StructureElement[] assertUpdateReject(StructureElementCommand[] structureElements) {
        return assertReject(assertUpdate(structureElements));
    }

    /**
     * Utility method to delete, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after delete)
     */
    public static StructureElement assertDeleteApprove(StructureElementCommand structureElement) {
        return assertApprove(assertDelete(structureElement));
    }
    /**
     * Utility method to delete, approve structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return approved structure elements (after delete)
     */
    public static StructureElement[] assertDeleteApprove(StructureElementCommand[] structureElements) {
        return assertApprove(assertDelete(structureElements));
    }

    /**
     * Utility method to delete, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after delete)
     */
    public static StructureElement assertDeleteCancel(StructureElementCommand structureElement) {
        return assertCancel(assertDelete(structureElement));
    }
    /**
     * Utility method to delete, cancel structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return cancelled structure elements (after delete)
     */
    public static StructureElement[] assertDeleteCancel(StructureElementCommand[] structureElements) {
        return assertCancel(assertDelete(structureElements));
    }

    /**
     * Utility method to delete, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after delete)
     */
    public static StructureElement assertDeleteReject(StructureElementCommand structureElement) {
        return assertReject(assertDelete(structureElement));
    }
    /**
     * Utility method to delete, reject structure elements and assert result.
     *
     * @param structureElements structure elements
     * @return rejected structure elements (after delete)
     */
    public static StructureElement[] assertDeleteReject(StructureElementCommand[] structureElements) {
        return assertReject(assertDelete(structureElements));
    }

}
