/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.junit.jupiter.api.Test;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.util.EnumUtil.StructureChoice;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is structures endpoint and discipline.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class StructuresDisciplineIT {

    // note
    //     StructureElement - when, who, comment - depend on requested or processed
    //         requested, requested by, requested comment
    //         processed, processed by, processed comment
    //     if less than a second between requested and processed, then considered one entry with processed
    //     history
    //         mnemonic path does not make same sense for history
    //         (very) tricky to find mnemonic path for uuid at proper time (history)
    //         therefore empty mnemonic path for history for structure
    //         one history entry if less than one second between requested and processed, otherwise two history entries
    //     attributes for entry for operations - create, update, delete, approve, cancel, reject
    //         some set client side, others set server side
    //         client side
    //             type, uuid, parent uuid, name, mnemonic, description, comment
    //         may be set client side for test purposes
    //     order of methods in test not known
    //         therefore mnemonic values chosen to not interfere with mnemonic values in other methods
    //
    //     StructureElement extends StructureElementCommand
    //     -->
    //     for such operations that send Command objects to backend, it is also possible to send StructureElement

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-integrationtest.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    @Test
    void equivalenceMnemonic() {
        // purpose
        //     test mnemonic equivalence
        //
        // what
        //     read        equivalence mnemonic

        ITUtilStructureElement.assertEquivalence("Di", "D1");
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create discipline
        //         not create itself
        //
        // what - combination of
        //     read        exists in structure
        //     read        is valid to create
        //     read        validate create
        //
        // validate create
        //     type
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            StructureElementCommand structureElement = new StructureElementCommand();

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "CC", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cc", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "CC", Boolean.TRUE);

            ITUtilStructureElement.assertValidate("[{asdf]", StructureChoice.CREATE, HttpURLConnection.HTTP_BAD_REQUEST);
            ITUtilStructureElement.assertCreate("[{asdf]", HttpURLConnection.HTTP_BAD_REQUEST);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElement.setType(Type.DISCIPLINE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElement.setName("name");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElement.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElement.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertCreate(structureElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            structureElement.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setType(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkCreateMnemonic() {
        // purpose
        //     test conditions for create discipline
        //         not create itself
        //
        // what - combination of
        //     read        validate create
        //
        // note
        //     mnemonic
        try {
            StructureElementCommand structureElement = new StructureElementCommand();

            structureElement.setType(Type.DISCIPLINE);
            structureElement.setName("name");
            structureElement.setDescription("description");
            structureElement.setComment("comment");

            // mnemonic rules

            structureElement.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("C");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Cc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Ccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Cccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Ccccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Cccccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Ccccccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Cccccccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Ccccccccc");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            // mnemonic rules (2)

            structureElement.setMnemonic(" ");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Dis ");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Dis");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("000");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("Dis0");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic(":");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Dis:");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("Dis:   ");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("1");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("12");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("123");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("1234");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("12345");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("123456");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);

            structureElement.setMnemonic("1234567");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("12345678");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);

            structureElement.setMnemonic("123456789");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createApprove() {
        // purpose
        //     test create and approve discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create in order to approve

        try {
            StructureElementCommand structureElement = null;
            StructureElement createdStructureElement = null;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Ca",
                    "description", "comment");

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Ca", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Ca", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Ca", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.TRUE);

            // approve
            ITUtilStructureElement.assertApprove(structureElement);

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Ca", Boolean.TRUE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Ca", Boolean.FALSE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createCancel() {
        // purpose
        //     test create and cancel discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate cancel
        //     patch       cancel structures
        //
        // note
        //     create in order to cancel

        try {
            StructureElementCommand structureElement = null;
            StructureElement createdStructureElement = null;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Cc",
                    "description", "comment");

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cc", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());
            UUID uuid = structureElement.getUuid();

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cc", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.TRUE);
            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);
            structureElement.setUuid(uuid);

            // cancel
            ITUtilStructureElement.assertCancel(structureElement);

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cc", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cc", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);
            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void createReject() {
        // purpose
        //     test create and reject discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate reject
        //     patch       reject structures
        //
        // note
        //     create in order to reject

        try {
            StructureElementCommand structureElement = null;
            StructureElement createdStructureElement = null;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Cr",
                    "description", "comment");

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cr", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());
            UUID uuid = structureElement.getUuid();

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cr", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.TRUE);
            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);
            structureElement.setUuid(uuid);

            // reject
            ITUtilStructureElement.assertReject(structureElement);

            ITUtilStructureElement.assertExists         (Type.DISCIPLINE, "Cr", Boolean.FALSE);
            ITUtilStructureElement.assertIsValidToCreate(Type.DISCIPLINE, "Cr", Boolean.TRUE);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);
            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CREATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update discipline
        //         not update
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update
        //
        // validate update
        //     uuid
        //     type
        //     name
        //     mnemonic
        //     description
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommand structureElement  = new StructureElementCommand();
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Cu",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());
            uuid = approvedStructureElement.getUuid();

            // validate update

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setType(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setType(Type.SYSTEMGROUP);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setType(Type.DISCIPLINE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setName(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setName("name");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setMnemonic("Cu");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setDescription(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);

            structureElement.setComment(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);

            structureElement.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateApprove() {
        // purpose
        //     test update and approve discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     update      update structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to update, approve

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Ua",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment update approve check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.TRUE);

            // approve
            ITUtilStructureElement.assertApprove(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateCancel() {
        // purpose
        //     test update and cancel discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate cancel
        //     update      update structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to update, cancel

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Uc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment update cancel check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.TRUE);

            // cancel
            ITUtilStructureElement.assertCancel(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void updateReject() {
        // purpose
        //     test update and reject discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate update
        //     read        validate approve
        //     read        validate reject
        //     update      update structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to update, reject

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Ur",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment update reject check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);

            // update
            ITUtilStructureElement.assertUpdate(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.TRUE);

            // reject
            ITUtilStructureElement.assertReject(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.UPDATE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete discipline
        //         not delete
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete
        //
        // validate delete
        //     uuid
        //     type
        //     comment

        try {
            UUID uuid = null;

            StructureElementCommand structureElement  = new StructureElementCommand();
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Cd",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());
            uuid = approvedStructureElement.getUuid();

            // validate delete

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setUuid(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);

            structureElement.setUuid(uuid);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setType(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);

            structureElement.setType(Type.SYSTEMGROUP);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);

            structureElement.setType(Type.DISCIPLINE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setName(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setName("name");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setMnemonic(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setMnemonic("Cd");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setDescription(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setDescription("description");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);

            structureElement.setComment(null);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);

            structureElement.setComment("comment");
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteApprove() {
        // purpose
        //     test delete and approve discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     delete      delete structures
        //     patch       approve structures
        //
        // note
        //     create, approve in order to delete, approve

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Da",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment delete approve check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE,  Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.TRUE);

            // approve
            ITUtilStructureElement.assertApprove(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE,  Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.APPROVE, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteCancel() {
        // purpose
        //     test delete and cancel discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate cancel
        //     update      delete structures
        //     patch       approve structures
        //     patch       cancel structures
        //
        // note
        //     create, approve in order to delete, cancel

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Dc",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment update delete check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.TRUE);

            // cancel
            ITUtilStructureElement.assertCancel(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.CANCEL, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void deleteReject() {
        // purpose
        //     test delete and reject discipline
        //
        // what - combination of
        //     create      create structures
        //     read        validate create
        //     read        validate delete
        //     read        validate approve
        //     read        validate reject
        //     update      delete structures
        //     patch       approve structures
        //     patch       reject structures
        //
        // note
        //     create, approve in order to delete, reject

        try {
            StructureElementCommand structureElement = null;
            StructureElement approvedStructureElement = null;

            // create, approve
            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Dr",
                    "description", "comment");
            approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
            structureElement.setUuid(approvedStructureElement.getUuid());

            structureElement.setComment("comment delete reject check");

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);

            // delete
            ITUtilStructureElement.assertDelete(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.FALSE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.TRUE);

            // reject
            ITUtilStructureElement.assertReject(structureElement);

            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.DELETE, Boolean.TRUE);
            ITUtilStructureElement.assertValidate(structureElement, StructureChoice.REJECT, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryApprove() {
        // purpose
        //     test read discipline in various ways for create, approve
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommand structureElement  = null;
            StructureElement createdStructureElement  = null;
            StructureElement approvedStructureElement = null;
            int length = 0;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Rsha",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());

            // read (1)
            length = ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rsha",   1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rsha",     0);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rsha", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rsha", Boolean.FALSE);

            // approve
            approvedStructureElement = ITUtilStructureElement.assertApprove(structureElement);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rsha",   0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rsha",  1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rsha", 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rsha",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + approvedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rsha",     1);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rsha", 1);

            ITUtilStructureElement.assertRead("/history/" + approvedStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rsha", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryCancel() {
        // purpose
        //     test read discipline in various ways for create, cancel
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommand structureElement   = null;
            StructureElement createdStructureElement   = null;
            StructureElement cancelledStructureElement = null;
            int length = 0;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Rshc",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());

            // read (1)
            length = ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rshc",   1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rshc", 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rshc", Boolean.FALSE);

            // cancel
            cancelledStructureElement = ITUtilStructureElement.assertCancel(structureElement);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rshc",   0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rshc", 1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rshc",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + cancelledStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshc",     0);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rshc", 0);

            ITUtilStructureElement.assertRead("/history/" + cancelledStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rshc", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryReject() {
        // purpose
        //     test read discipline in various ways for create, reject
        //         search
        //         latest
        //         history
        //
        // note
        //     create (and more) to read (with content)

        try {
            StructureElementCommand structureElement  = null;
            StructureElement createdStructureElement  = null;
            StructureElement rejectedStructureElement = null;
            int length = 0;

            structureElement = new StructureElementCommand(
                    null, Type.DISCIPLINE, null,
                    "name", "Rshr",
                    "description", "comment");

            // create
            createdStructureElement = ITUtilStructureElement.assertCreate(structureElement);
            structureElement.setUuid(createdStructureElement.getUuid());

            // read (1)
            length = ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize();

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rshr",   1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + createdStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + createdStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rshr", Boolean.FALSE);

            // reject
            rejectedStructureElement = ITUtilStructureElement.assertReject(structureElement);

            // read (2)
            assertEquals(length, ITUtilStructureElement.assertRead("/DISCIPLINE", 1, -1).getListSize());

            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING",   0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED",  0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED", 0, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED",  1, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=Rshr",   0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=Rshr",  0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=Rshr", 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=Rshr",  1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + rejectedStructureElement.getUuid().toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/Rshr",     0);

            ITUtilStructureElement.assertRead("/mnemonicpath/Rshr", 0);

            ITUtilStructureElement.assertRead("/history/" + rejectedStructureElement.getUuid().toString(), 1, -1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "Rshr", Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest1() {
        // purpose
        //     test read discipline in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         1 entry for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommand structureElement  = null;
        StructureElement responseStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "P1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);
        uuid = responseStructureElement.getUuid();

        // a number of entries for which the last entry has status CANCELLED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "C1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);
        uuid2 = responseStructureElement.getUuid();

        // a number of entries for which the last entry has status REJECTED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "R1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);
        uuid3 = responseStructureElement.getUuid();

        // a number of entries for which the last entry has status APPROVED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "A1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid4 = responseStructureElement.getUuid();

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid2.toString(), 1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid3.toString(), 1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/C1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/R1", 0);
            ITUtilStructureElement.assertRead("/mnemonic/A1", 1);

            // normally expected number of entries are 1, 2, 2, 2
            // since within 1 second from requested to processed, expected number of entries 1, 1, 1, 1
            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  1);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 1);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 1);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "P1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "C1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "R1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "A1", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryStatusLatest9() {
        // purpose
        //     test read discipline in various ways
        //         status
        //         latest
        //       ( deleted )
        //         history
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     4 different lines  of uuid with combinations of status and latest
        //         9 entries for each line of uuid
        //         in particular different statuses for last entry in lines of uuid
        //         mnemonic - P, C, R, A (pending, cancelled, rejected, approved)
        //
        // note!
        //     not only
        //         exclude APPROVED and not latest
        //     not only
        //         exclude earlier than APPROVED and latest (considered history)
        //         include PENDING and not latest           (considered future)
        //     instead
        //         exclude earlier than APPROVED and latest (considered history)
        //       ( include if latest )
        //         include if latest not exists             (considered present + future)

        StructureElementCommand structureElement  = null;
        StructureElement responseStructureElement = null;
        UUID uuid, uuid2, uuid3, uuid4 = null;

        // a number of entries for which the last entry has status PENDING
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "P9",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid = responseStructureElement.getUuid();

        responseStructureElement.setComment("comment 2");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 3");
        responseStructureElement = ITUtilStructureElement.assertUpdateCancel(responseStructureElement);
        responseStructureElement.setComment("comment 4");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 5");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 6");
        responseStructureElement = ITUtilStructureElement.assertUpdateReject(responseStructureElement);
        responseStructureElement.setComment("comment 7");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 8");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 9");
        responseStructureElement = ITUtilStructureElement.assertUpdate(responseStructureElement);

        // a number of entries for which the last entry has status CANCELLED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "C9",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid2 = responseStructureElement.getUuid();

        responseStructureElement.setComment("comment 2");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 3");
        responseStructureElement = ITUtilStructureElement.assertUpdateCancel(responseStructureElement);
        responseStructureElement.setComment("comment 4");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 5");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 6");
        responseStructureElement = ITUtilStructureElement.assertUpdateReject(responseStructureElement);
        responseStructureElement.setComment("comment 7");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 8");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 9");
        responseStructureElement = ITUtilStructureElement.assertUpdateCancel(responseStructureElement);

        // a number of entries for which the last entry has status REJECTED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "R9",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid3 = responseStructureElement.getUuid();

        responseStructureElement.setComment("comment 2");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 3");
        responseStructureElement = ITUtilStructureElement.assertUpdateCancel(responseStructureElement);
        responseStructureElement.setComment("comment 4");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 5");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 6");
        responseStructureElement = ITUtilStructureElement.assertUpdateReject(responseStructureElement);
        responseStructureElement.setComment("comment 7");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 8");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 9");
        responseStructureElement = ITUtilStructureElement.assertUpdateReject(responseStructureElement);

        // a number of entries for which the last entry has status APPROVED
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "A9",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid4 = responseStructureElement.getUuid();

        responseStructureElement.setComment("comment 2");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 3");
        responseStructureElement = ITUtilStructureElement.assertUpdateCancel(responseStructureElement);
        responseStructureElement.setComment("comment 4");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 5");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 6");
        responseStructureElement = ITUtilStructureElement.assertUpdateReject(responseStructureElement);
        responseStructureElement.setComment("comment 7");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 8");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);
        responseStructureElement.setComment("comment 9");
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(responseStructureElement);

        try {
            // from first structure element
            assertNotNull(uuid);

            // note
            //     exclude earlier than APPROVED and latest (considered history)
            //   ( include if latest )
            //     include if latest not exists             (considered present + future)

            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid.toString(),  2);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid2.toString(), 2);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid3.toString(), 2);
            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid4.toString(), 1);

            ITUtilStructureElement.assertRead("/mnemonic/P9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/C9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/R9", 1);
            ITUtilStructureElement.assertRead("/mnemonic/A9", 1);

            // normally expected number of entries are 17, 18, 18, 18
            // since within 1 second from requested to processed, expected number of entries 9, 9, 9, 9
            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  9);
            ITUtilStructureElement.assertRead("/history/" + uuid2.toString(), 9);
            ITUtilStructureElement.assertRead("/history/" + uuid3.toString(), 9);
            ITUtilStructureElement.assertRead("/history/" + uuid4.toString(), 9);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "P9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "C9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "R9", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "A9", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchStatusDeletedChildren() {
        // purpose
        //     test read discipline in various ways
        //         status
        //       ( latest )
        //         deleted
        //         children
        //
        // what
        //     entries with different statuses
        //
        // note
        //     create (and more) to read (with content)
        //     querying for Status.APPROVED means latest approved

        StructureElementCommand structureElement  = null;
        StructureElement responseStructureElement = null;
        UUID uuid = null;

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AA1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        uuid = responseStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AA2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AA3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AA4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AA5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AB1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AB2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AB3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AB4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AB5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateCancel(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AC1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AC2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AC3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AC4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AC5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AD1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AD2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AD3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AD4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);
        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AD5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreate(structureElement);

        String description2 = "some other description";
        String comment2 = "some other comment";
        String description3 = "more description";
        String comment3 = "more comment";

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AE1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description2);
        structureElement.setComment(comment2);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description3);
        structureElement.setComment(comment3);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AE2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description2);
        structureElement.setComment(comment2);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description3);
        structureElement.setComment(comment3);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AE3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description2);
        structureElement.setComment(comment2);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description3);
        structureElement.setComment(comment3);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AE4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description2);
        structureElement.setComment(comment2);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description3);
        structureElement.setComment(comment3);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AE5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description2);
        structureElement.setComment(comment2);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        structureElement.setDescription(description3);
        structureElement.setComment(comment3);
        responseStructureElement = ITUtilStructureElement.assertUpdateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteReject(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AF1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AF2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AF3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AF4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AF5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDeleteApprove(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AG1",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDelete(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AG2",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDelete(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AG3",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDelete(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AG4",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDelete(structureElement);

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "name", "AG5",
                "description", "comment");
        responseStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        structureElement = responseStructureElement;
        responseStructureElement = ITUtilStructureElement.assertDelete(structureElement);

        // 60 discipline entries

        try {
            // from first structure element
            assertNotNull(uuid);

            // read & search
            //     /{type}
            //     type, statuses, deleted, uuid, parentuuid, name, mnemonic, mnemonicequivalence, mnemonicpath, description
            //     combination
            ITUtilStructureElement.assertRead("/DISCIPLINE?mnemonic=A__",                                                  45);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&mnemonic=A__",                                 10);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=APPROVED&mnemonic=A__",                                20);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=CANCELLED&mnemonic=A__",                                5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=REJECTED&mnemonic=A__",                                10);
            ITUtilStructureElement.assertRead("/DISCIPLINE?statuses=PENDING&statuses=APPROVED&mnemonic=A__",               30);

            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&mnemonic=A__",                                    30);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&statuses=PENDING&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&statuses=APPROVED&mnemonic=A__",                  15);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&statuses=CANCELLED&mnemonic=A__",                  5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&statuses=REJECTED&mnemonic=A__",                   5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=false&statuses=PENDING&statuses=APPROVED&mnemonic=A__", 20);

            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&mnemonic=A__",                                     15);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&statuses=PENDING&mnemonic=A__",                     5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&statuses=APPROVED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&statuses=CANCELLED&mnemonic=A__",                   0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&statuses=REJECTED&mnemonic=A__",                    5);
            ITUtilStructureElement.assertRead("/DISCIPLINE?deleted=true&statuses=PENDING&statuses=APPROVED&mnemonic=A__",  10);

            ITUtilStructureElement.assertRead("/DISCIPLINE?uuid=" + uuid.toString(),        1);

            ITUtilStructureElement.assertRead("/DISCIPLINE?parentuuid=" + uuid.toString(),  0);

            ITUtilStructureElement.assertRead("/DISCIPLINE?name=na",                        0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?name=na_",                       0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?name=na__",                     45, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?name=na___",                     0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?name=na%",                      45, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?name=name",                     45, -1);

            ITUtilStructureElement.assertRead("/DISCIPLINE?mnemonicequivalence=A__",       45, -1);

            ITUtilStructureElement.assertRead("/DISCIPLINE?mnemonicepath=A__",             45, -1);

            ITUtilStructureElement.assertRead("/DISCIPLINE?description=desc",               0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=desc%",             35, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=sc",                 0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=sc%",                0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=%sc",                0);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=%sc%",              45, -1);
            ITUtilStructureElement.assertRead("/DISCIPLINE?description=description",       35, -1);

            // children
            //     /children/{uuid}
            //     uuid, type, statuses, deleted, name, mnemonic, mnemonicequivalence, mnemonicpath, description
            ITUtilStructureElement.assertRead("/children/" + uuid.toString(),                      0);
            ITUtilStructureElement.assertRead("/children/" + uuid.toString() + "?type=DISCIPLINE", 0);

            // mnemonic
            //     /mnemonic/{mnemonic}
            ITUtilStructureElement.assertRead("/mnemonic/A",        0);
            ITUtilStructureElement.assertRead("/mnemonic/A__",     15, -1);
            ITUtilStructureElement.assertRead("/mnemonic/AG_",      5);
            ITUtilStructureElement.assertRead("/mnemonic/AG1",      1);

            // mnemonic path
            //     /mnemonicpath/{mnemonicpath}
            ITUtilStructureElement.assertRead("/mnemonicpath/A",    0);
            ITUtilStructureElement.assertRead("/mnemonicpath/A__", 15, -1);
            ITUtilStructureElement.assertRead("/mnemonicpath/AG_",  5);
            ITUtilStructureElement.assertRead("/mnemonicpath/AG1",  1);

            // history
            //     /history/{uuid}
            ITUtilStructureElement.assertRead("/history/" + uuid.toString(),  1);

            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AA1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AA2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AA3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AA4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AA5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AB1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AB2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AB3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AB4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AB5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AC1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AC2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AC3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AC4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AC5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AD1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AD2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AD3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AD4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AD5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AE1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AE2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AE3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AE4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AE5", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AF1", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AF2", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AF3", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AF4", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AF5", Boolean.FALSE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AG1", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AG2", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AG3", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AG4", Boolean.TRUE);
            ITUtilStructureElement.assertExists(Type.DISCIPLINE, "AG5", Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

}
